/**
 * Title:        Channel models
 * Description:  AWGN and BSC channel models
 * Copyright:    Copyright (c) 2001, Ling Li
 * Company:      California Institute of Technology
 * @author Ling Li
 * @email  ling@cs.caltech.edu
 * @version 1.0
 */

package ecc;
import java.lang.Math;

public class Channel {
    private Random rnd;
    private double n1, n2;

    public Channel() {
        rnd = new Random();
        setSeed(100);
    }

    public Channel(int seed) {
        rnd = new Random();
        setSeed(seed);
    }

    public void setSeed(int seed) {
        rnd.setSeed(seed);
    }

    public int getSeed() {
        return rnd.getSeed();
    }

    /** AWGN
     *  with binary inputs {0, 1}, return output of an AWGN channel.
     *  r[i] is a vector of length 2 in {-1, +1}
     */
    public double[][] AWGN(int[][] y, double sigma) {
        int len = y.length;
        double r[][] = new double[len][2];

        for (int i = 0; i < len; i++) {
            double randn[] = normal(sigma);
            r[i][0] = 1 - 2 * y[i][0] + randn[0];
            r[i][1] = 1 - 2 * y[i][1] + randn[1];
        }

        return r;
    }

    /** AWGN2BSC
     *  change vectors in {-1, +1} to {0, 1}
     */
    public int[][] AWGN2BSC(double[][] r) {
        int len = r.length;
        int[][] y = new int[len][2];

        for (int i = 0; i < len; i++) {
            y[i][0] = (r[i][0] < 0)? 1 : 0;
            y[i][1] = (r[i][1] < 0)? 1 : 0;
        }

        return y;
    }

    private double[] normal(double sigma) {
        double x1, x2, s;
        do {
            x1 = rnd.urand1();
            x2 = rnd.urand1();
            x1 = 2 * x1 - 1;
            x2 = 2 * x2 - 1;

            s = x1 * x1 + x2 * x2;
        } while (s >= 1.0);
        s = sigma * Math.sqrt(-2 * Math.log(s) / s);

        double n[] = new double[2];
        n[0] = x1 * s;
        n[1] = x2 * s;
        return n;
    }
}
