/**
 * Title:        Turbo Decoder for convolutional codes
 * Description:  An abstract TurboCode class
 * Copyright:    Copyright (c) 2001, Ling Li
 * Company:      California Institute of Technology
 * @author Ling Li
 * @email  ling@cs.caltech.edu
 * @version 1.1
 */

package ecc;

abstract public class TurboCode {
    final int mem = 4;
    final int states = 1 << mem;
    final int state_mask = states-1;

    // trellis[st][x] is the next state from st with input x.
    // invtrel[st][x] is the previous state with input x ending at st.
    private int trellis[][], invtrel[][];
    // outputs(st, x) is the output at state st with input x
    private int output0[][], output1[][];

    private double LLR[];

    private void make_trellis() {
        for (int st = 0; st < states; st++) {
            // st >> i is D^(i+1)
            int sumst = 1 & (st ^ (st >> 1) ^ (st >> 2) ^ (st >> 3));
            int newst = ((st << 1) & state_mask) | sumst;
            sumst ^= (st >> 3);  // 1 & (st >> 3)
            for (int bit = 0; bit < 2; bit++) {
                trellis[st][bit] = newst ^ bit;
                invtrel[newst ^ bit][bit] = st;
                output0[st][bit] = bit;
                output1[st][bit] = bit ^ sumst;
            }
        }
    }

    public TurboCode() {
        trellis = new int[states][2];
        output0 = new int[states][2];
        output1 = new int[states][2];
        invtrel = new int[states][2];

        make_trellis();
    }

    /** y = encode(x);
     *  where x is a row vector containing information
     *  we will try, after trunc_len dummy bits, to ensure the state be 0.
     */
    public int[][] encode(int[] x, int trunc_len) {
        int len = x.length, state, i;
        int y[][];

        y = new int[len+trunc_len][2];
        state = 0;
        for (i = 0; i < len; i++) {
            y[i][0] = output0[state][x[i]];
            y[i][1] = output1[state][x[i]];
            state   = trellis[state][x[i]];
        }

        for (i = 0; i < trunc_len; i++) {
            int xr = 1 & (state ^ (state >> 1) ^ (state >> 2) ^ (state >> 3));
            y[len+i][0] = output0[state][xr];
            y[len+i][1] = output1[state][xr];
            state   = trellis[state][xr];
        }

        return y;
    }

    abstract protected double logsum(double x, double y);

    public int[] decode(double[][] r, int trunc_len, double sigma, double p0) {
        int len = r.length;
        if (len <= trunc_len) return null;
        int x[] = new int[len - trunc_len];

        double lp0 = Math.log(p0), lp1 = Math.log(1-p0);
        double snr = 1 / sigma / sigma;

        double beta[][]  = new double[len+1][states];
        int s, s0, s1, t;

        beta[len][0] = 0;
        for (s = 1; s < states; s++)
            beta[len][s] = Double.NEGATIVE_INFINITY;

        for (t = len; t > 0; t--) {
            for (s = 0; s < states; s++) {
                double d0, d1, w0, w1;

                s0 = trellis[s][0];
                d0 = r[t-1][0] * (1-2*output0[s][0]);
                d1 = r[t-1][1] * (1-2*output1[s][0]);
                w0 = lp0 + snr * (d0 + d1);

                s1 = trellis[s][1];
                d0 = r[t-1][0] * (1-2*output0[s][1]);
                d1 = r[t-1][1] * (1-2*output1[s][1]);
                w1 = lp1 + snr * (d0 + d1);

                beta[t-1][s] = logsum(beta[t][s0] + w0, beta[t][s1] + w1);
            }
        }

        double alpha[] = new double[states];
        double alpha_new[] = new double[states];
        alpha[0] = 0;
        for (s = 1; s < states; s++)
            alpha[s] = Double.NEGATIVE_INFINITY;

        // Calculate alpha & LLR
        LLR = new double[len];  // in fact, len - trunc_len ???
        double aw[][] = new double[states][2];
        for (t = 0; t < len - trunc_len; t++) {
            double pu0, pu1;
            double d0, d1;
            pu0 = pu1 = Double.NEGATIVE_INFINITY;
            for (s = 0; s < states; s++) {
                s0 = trellis[s][0];
                d0 = r[t][0] * (1-2*output0[s][0]);
                d1 = r[t][1] * (1-2*output1[s][0]);
                aw[s][0] = alpha[s] + lp0 + snr * (d0 + d1);

                s1 = trellis[s][1];
                d0 = r[t][0] * (1-2*output0[s][1]);
                d1 = r[t][1] * (1-2*output1[s][1]);
                aw[s][1] = alpha[s] + lp1 + snr * (d0 + d1);

                pu0 = logsum(pu0, aw[s][0] + beta[t+1][s0]);
                pu1 = logsum(pu1, aw[s][1] + beta[t+1][s1]);
            }
            LLR[t] = pu0 - pu1;
            x[t] = (LLR[t] < 0)? 1 : 0;

            for (s = 0; s < states; s++) {
                s0 = invtrel[s][0];
                s1 = invtrel[s][1];

                alpha_new[s] = logsum(aw[s0][0], aw[s1][1]);
            }
            double[] temp = alpha;
            alpha = alpha_new;
            alpha_new = temp;
        }

        return x;
    }

    public double[] getLLR() {
        return LLR;
    }
}